//!
//!
use super::{models, schema::member_bank_accounts};
use beteran_common_rust as bcr;
use diesel::prelude::*;
use diesel::result::Error;

///
pub struct Repository {}

impl std::fmt::Debug for Repository {
  fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
    f.debug_struct("Repository of member_bank_accounts")
      .finish()
  }
}

impl Default for Repository {
  fn default() -> Self {
    Self::new()
  }
}

impl Repository {
  ///
  pub fn new() -> Repository {
    Repository {}
  }

  ///
  pub fn insert(
    &self,
    conn: &diesel::PgConnection,
    new_resource: &models::NewMemberBankAccount,
  ) -> Result<models::MemberBankAccount, Error> {
    let inserted = diesel::insert_into(member_bank_accounts::table)
      .values(new_resource)
      .get_result::<models::MemberBankAccount>(conn)?;

    Ok(inserted)
  }

  ///
  pub fn select(
    &self,
    conn: &diesel::PgConnection,
    id: uuid::Uuid,
  ) -> Result<models::MemberBankAccount, Error> {
    member_bank_accounts::table
      .find(id as uuid::Uuid)
      .first::<models::MemberBankAccount>(conn)
  }

  ///
  pub fn select_all_count(
    &self,
    conn: &diesel::PgConnection,
    find_all: models::FindAll,
  ) -> Result<i64, Error> {
    let mut q = member_bank_accounts::table.into_boxed();

    if let Some(sp) = find_all.member_id {
      q = q.filter(member_bank_accounts::dsl::member_id.eq(sp));
    }
    if let Some(sp) = find_all.bank_id {
      q = q.filter(member_bank_accounts::dsl::bank_id.eq(sp));
    }
    if let Some(sp) = find_all.name_like {
      q = q.filter(member_bank_accounts::dsl::name.like(sp));
    }
    if let Some(sp) = find_all.account_number_like {
      q = q.filter(member_bank_accounts::dsl::account_number.like(sp));
    }
    if let Some(sp) = find_all.memo_like {
      q = q.filter(member_bank_accounts::dsl::memo.like(sp));
    }

    q.count().get_result(conn)
  }

  ///
  pub fn select_all(
    &self,
    conn: &diesel::PgConnection,
    find_all: models::FindAll,
  ) -> Result<Vec<models::MemberBankAccount>, Error> {
    let mut q = member_bank_accounts::table.into_boxed();

    if let Some(sp) = find_all.member_id {
      q = q.filter(member_bank_accounts::dsl::member_id.eq(sp));
    }
    if let Some(sp) = find_all.bank_id {
      q = q.filter(member_bank_accounts::dsl::bank_id.eq(sp));
    }
    if let Some(sp) = find_all.name_like {
      q = q.filter(member_bank_accounts::dsl::name.like(sp));
    }
    if let Some(sp) = find_all.account_number_like {
      q = q.filter(member_bank_accounts::dsl::account_number.like(sp));
    }
    if let Some(sp) = find_all.memo_like {
      q = q.filter(member_bank_accounts::dsl::memo.like(sp));
    }

    if let Some(p) = find_all.pagination {
      let page = p.page.unwrap_or(1);

      if let Some(page_size) = p.page_size {
        q = q.offset(((page - 1) * page_size) as i64);
        q = q.limit(page_size as i64);
      }
    }
    if let Some(orderbys) = find_all.sorts {
      for s in orderbys {
        match s {
          bcr::models::pagination::Sort::ASC(property) => match property.as_str() {
            "name" => {
              q = q.order_by(member_bank_accounts::name.asc());
            }
            "account_number" => {
              q = q.order_by(member_bank_accounts::account_number.asc());
            }
            "created_at" => {
              q = q.order_by(member_bank_accounts::created_at.asc());
            }
            "updated_at" => {
              q = q.order_by(member_bank_accounts::updated_at.asc());
            }
            _ => {}
          },
          bcr::models::pagination::Sort::DESC(property) => match property.as_str() {
            "name" => {
              q = q.order_by(member_bank_accounts::name.desc());
            }
            "account_number" => {
              q = q.order_by(member_bank_accounts::account_number.desc());
            }
            "created_at" => {
              q = q.order_by(member_bank_accounts::created_at.desc());
            }
            "updated_at" => {
              q = q.order_by(member_bank_accounts::updated_at.desc());
            }

            _ => {}
          },
        };
      }
    }

    q.load::<models::MemberBankAccount>(conn)
  }

  ///
  pub fn update(
    &self,
    conn: &diesel::PgConnection,
    id: uuid::Uuid,
    modify: &models::ModifyMemberBankAccount,
  ) -> Result<u64, Error> {
    use member_bank_accounts::dsl;

    diesel::update(dsl::member_bank_accounts.filter(dsl::id.eq(id)))
      .set(modify)
      .execute(conn)
      .map(|c| c as u64)
  }
}
